from __future__ import annotations

from os import fdopen as os_fdopen
from typing import TYPE_CHECKING

from plyer.utils import platform

if platform == 'android':
	from .android_objects import ContentProvider, ContentResolver, DCDocument, DocumentsContract

from .io import async_open, wrap2async
from .utils import generate_file_uri_from_access_uri, get_fd_from_android_uri, scoped_res_exists

if TYPE_CHECKING:
	from io import BufferedReader, BufferedWriter, TextIOWrapper

	from aiofiles.threadpool.binary import AsyncBufferedIOBase, AsyncBufferedReader
	from aiofiles.threadpool.text import AsyncTextIOWrapper


def scoped_file_open(
	access_uri: 'jni[android.net.Uri]',
	name: str,
	mode: str = 'w',
	mime: str = '*/*',
) -> int:
	"""Make file using access (usually directory) android uri and return the file descriptor."""
	# TODO: Optionally return file uri
	file_uri: 'jni[android.net.Uri]' = generate_file_uri_from_access_uri(access_uri, name)
	if scoped_res_exists(file_uri):
		if mode == 'x':
			raise FileExistsError

	elif mode in ('r', 'rb', 'r+', 'rb+'):
		raise FileNotFoundError

	else:
		del file_uri
		# "w/wb", "a/ab", "r+/rb+"
		# "a/ab" - because usually it's makes new file if not already exists to append
		file_uri = dc_make_doc(
			access_uri,
			name, mime,
		)


	ret = get_fd_from_android_uri(file_uri, mode)

	return ret


def dc_make_doc(
	access_uri: 'jni[android.net.Uri]',
	name: str,
	mime: str = '*/*',
) -> 'jni[android.net.Uri]':
	"""Make file using access (usually directory) android uri and return the file uri."""
	# TODO: Optionally return file uri
	# TODO: Raise error if file exists..
	return DocumentsContract.createDocument(
		ContentResolver,
		access_uri,
		mime, name,
	)


def dc_open_doc(
	access_uri: 'jni[android.net.Uri]',
	name: str,
) -> 'jni[android.net.Uri]':
	"""Open file using access (usually directory) android uri and return the file uri."""
	# TODO: Optionally return file uri
	# FIXME: Raise error if file exists..
	return ContentProvider.createDocument(
		ContentResolver,
		access_uri,
		mime, name,
	)


def scoped_file_open_sync(
	access_uri: 'android.net.Uri',
	name: str,
	mode: str = 'w',
	mime: str = '*/*',
) -> TextIOWrapper | BufferedReader | BufferedWriter:
	return os_fdopen(scoped_file_open(access_uri, name, mode, mime), mode)


def scoped_file_open_async(
	access_uri: 'android.net.Uri',
	name: str,
	mode: str = 'w',
	mime: str = '*/*',
) -> AsyncTextIOWrapper | AsyncBufferedReader | AsyncBufferedIOBase:
	return async_open(scoped_file_open(access_uri, name, mode, mime), mode)


class DirectoryExistsError(FileExistsError):
	pass


def scoped_mkdir_sync(
	access_uri: 'jni[android.net.Uri]',
	name: str,
	*,
	exist_ok: bool = False,
) -> 'jni[android.net.Uri]':
	# Why Google doesn't made separate method for this stuff?
	uri = DocumentsContract.createDocument(
		ContentResolver,
		access_uri, DCDocument.MIME_TYPE_DIR, name,
	)
	if scoped_res_exists(uri) and not exist_ok:
		# TODO: More informative message..
		raise DirectoryExistsError
	return uri


sfopen_sync = scoped_file_open_sync
sfopen_async = scoped_file_open_async

smkdir_sync = scoped_mkdir_sync
scoped_mkdir_async = wrap2async(scoped_mkdir_sync)
smkdir_async = scoped_mkdir_async



from __future__ import annotations

from typing import TYPE_CHECKING

from plyer.utils import platform

if platform == 'android':
	from jnius.jnius import JavaException

	from .android_objects import ContentResolver, DCDocument, DocumentsContract, Uri

PYSLET_AVAILABLE: bool = True

try:
	from pyslet.rfc2396 import URI
except ImportError:
	PYSLET_AVAILABLE = False

if TYPE_CHECKING:
	from typing import Dict, Final  # noqa: UP035

	from pyslet.rfc2396 import URI


def scoped_res_exists(file_uri: 'jni[android.net.Uri]') -> bool:
	c: 'jni[android.database.Cursor] | None' = None
	try:
		c = ContentResolver.query(
			file_uri,
			(DCDocument.COLUMN_DOCUMENT_ID,),
			# [DCDocument.COLUMN_DOCUMENT_ID],
			None, None, None,
		)
		return c.getCount() > 0
	except JavaException as e:
		if e.innermessage is None or 'java.io.FileNotFoundException' in e.innermessage:
			# FIXME: Log fail..
			del e
			return False

		raise e

	# All's fine
	# TODO: Quite close like in `DocumentsContractApi19.java`
	if c is not None:
		c.close()

	return True


def _old_scoped_file_exists(file_uri: 'jni[android.net.Uri]') -> bool:
	try:
		ins = ContentResolver.openInputStream(file_uri)
	except JavaException as e:
		if 'java.io.FileNotFoundException' in e.innermessage:
			del e
			return False

		raise e

	# All's fine
	ins.close()
	del ins

	return True


# TODO: Use enums..
# NOTE: java "w" -> "wt", "rw" -> "rwt" because: https://issuetracker.google.com/issues/180526528
py_open2java_openfd: 'Final[Dict[str, str]]' = {  # noqa: UP006,UP037
	'r': 'r',
	'rb': 'r',

	# TODO: Optionally skip truncate..
	# java "w" -> "wt"
	'w': 'wt',
	'wb': 'wt',
	'w+': 'wt',
	'wb+': 'wt',

	'x': 'wt',  # existent check is inside sfopen_{sync/async}
	'xb': 'wt',

	'a': 'wa',
	'ab': 'wa',

	'a+': 'rwa',
	'ab+': 'rwa',

	# no truncate, because
	# usually in stdlib `io.open` if file will just read, it's doesn't be overwritten
	'r+': 'rw',
	'rb+': 'rw',
}


# TODO: Handle access errors..
def get_fd_from_android_uri(
	content_uri: 'android.net.Uri',
	mode: str = 'r',
) -> int:
	"""Open and detach android java file descriptor for	directly access in python.

	Note for open modes:
	https://developer.android.com/reference/android/content/ContentResolver#openFileDescriptor(android.net.Uri,%20java.lang.String,%20android.os.CancellationSignal)
	"""
	# TODO: CancellationSignal..

	# Stuff around modes to correctly use in the java descriptor
	# TODO: Raise another error..
	jp_mode = py_open2java_openfd[mode]

	fd_obj: 'jni[android.os.ParcelFileDescriptor]' = ContentResolver.openFileDescriptor(
		content_uri,
		jp_mode,
	)
	fd: int = fd_obj.detachFd()

	return fd


def get_fd_from_uri(
	uri: URI,
	mode: str = 'r',
) -> int:
	android_uri: 'android.net.Uri' = Uri.parse(str(uri))
	return get_fd_from_android_uri(android_uri, mode)


def get_fd_from_struri(
	struri: str,
	mode: str = 'r',
) -> int:
	android_uri: 'android.net.Uri' = Uri.parse(struri)
	return get_fd_from_android_uri(android_uri, mode)


# TODO: Speed check..
# Weak.. Do we really need pyslet's rfc2396?
# def _pyslet_generate_file_uri_from_access_uri(
	# access_uri: 'jni[android.net.Uri]',
# 	name: str,
# ) -> 'jni[android.net.Uri]':
# 	py_uri: URI = URI(access_uri.toString())
# 	doc_file: str = f'{py_uri.get_file_name()}/{name}'

# 	# FIXME: Find better ways to natively use rfc2396 Uri..
# 	# py_file_uri: str = str(py_uri.resolve(doc_file))##
# 	# ret = Uri.parse(py_file_uri)

# 	ret = DocumentsContract.buildDocumentUriUsingTree(access_uri, doc_file)

# 	del py_uri, doc_file

# 	return ret


# TODO: Handle parse errors..
def generate_file_uri_from_access_uri(access_uri: 'jni[android.net.Uri]', name: str) -> 'jni[android.net.Uri]':
	dir_name: str = DocumentsContract.getTreeDocumentId(access_uri)
	doc_file: str = f'{dir_name}/{name}'
	del dir_name

	ret = DocumentsContract.buildDocumentUriUsingTree(access_uri, doc_file)

	del doc_file

	return ret


# TODO: Optionally allow force use java api more..
# TODO: More use `from functools import singledispatch`
# generate_file_uri_from_access_uri = _java_generate_file_uri_from_access_uri

# if PYSLET_AVAILABLE:
# 	generate_file_uri_from_access_uri = _pyslet_generate_file_uri_from_access_uri



from __future__ import annotations

from plyer.utils import platform

if platform != 'android':
	msg = "Can't access android api"
	raise AttributeError(msg)

from android import mActivity
from jnius import autoclass
from plyer.platforms.android.filechooser import DocumentsContract, Intent  # noqa: F401

ContentResolver = mActivity.getContentResolver()
Uri = autoclass('android.net.Uri')

# TODO: Try to replace `DocumentsContract`
ContentProvider = autoclass('android.content.ContentProvider')

DCDocument = autoclass('android.provider.DocumentsContract$Document')



from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from aiofiles import open as async_open
else:
	def _throw_plug(p, m, **kw):
		raise ModuleNotFoundError

try:
	from aiofiles import open as async_open
	from aiofiles.ospath import wrap as wrap2async
except ImportError:
	async_open = _throw_plug
	wrap2async = _throw_plug



from __future__ import annotations

from typing import TYPE_CHECKING

from plyer import filechooser
from plyer.utils import platform

if platform == 'android':
	from functools import partial

	from android import activity

	from .android_objects import Intent, mActivity

if TYPE_CHECKING:
	from typing import Any

	from .android_objects import Uri


# TODO: Just extend class..

# FIXME: Look at mypy docs & annotate more better..
def pyss_resolve_uri(uri: list[Uri] | Uri) -> list[Uri] | Uri:
	"""Append a few paths to dir uri.

	For correctly work with `PyScopedStorage.tools.mksf_{sync,async}`.
	"""
	filechooser.selection_raw = uri
	_uri: Uri = uri
	is_uri_list = False
	if isinstance(uri, list):
		is_uri_list = True
		_uri = uri[0]

	# TODO: More tests..
	# NOTE: More use `pyslet.rfc2396.URI` in your python code
	ret = _uri.buildUpon().appendPath('document').appendPath(_uri.getLastPathSegment())
	if is_uri_list:
		assert not isinstance(_uri, list)  # linter plug
		_uri[0] = ret
		return _uri

	return ret


def pyss_open_dir(**kwargs: dict[str, Any]) -> None:
	filechooser._handle_selection = kwargs.pop(  # noqa: SLF001
		'on_selection', filechooser._handle_selection,  # noqa: SLF001
	)

	intent: 'jni[android.content.Intent]' = Intent(Intent.ACTION_OPEN_DOCUMENT_TREE)

	# TODO: Handle persistent req

	mActivity.startActivityForResult(intent, filechooser.select_code)

	del intent


def pyss_on_activity_result(self, request_code: int, result_code: int, data: Intent) -> None:
	sel = self.selection_raw if not isinstance(self.selection_raw, list) else self.selection_raw[0]
	# FIXME: Clean `self.selection_raw`
	# TODO: Ignore on file calls..?
	resolver: 'jni[android.content.ContentResolver]' = mActivity.getContentResolver()
	resolver.takePersistableUriPermission(
		sel,
		# TODO: Make flags as option..
		Intent.FLAG_GRANT_READ_URI_PERMISSION | Intent.FLAG_GRANT_WRITE_URI_PERMISSION,
	)

	del resolver


def pyss_file_selection_dialog(**kwargs: dict[str, Any]) -> None:
	"""Open selection dialog with dir select support."""
	mode = kwargs.pop('mode', None)
	if mode == 'open':
		filechooser._open_file(**kwargs)  # noqa: SLF001

	if mode == 'dir':
		pyss_open_dir(**kwargs)


if platform == 'android':
	# We're processing uris using `PyScopedStorage.tools.make_scoped_file` & we don't need this method
	filechooser._resolve_uri = pyss_resolve_uri  # noqa: SLF001

	# Override filechooser instance method to handle directory choose
	filechooser._file_selection_dialog = pyss_file_selection_dialog  # noqa: SLF001
	# Bind take persistent uri permission to dir selection
	# (you can get it using ContentResolver's method `getPersistedUriPermissions`)
	activity.bind(on_activity_result=partial(pyss_on_activity_result, filechooser))

	# NOTE: activity.bind binds on object pointer itself, not on the name
	# (hold it in the head if you use class methods or etc.)
